import { URLChangeObservable } from '../../lib/detect-url-change';
import { logger } from '../../lib/logger';
import { createInteractionSignal, createNavigationSignal } from '../../types';
const parseLabels = (labels) => {
    if (!labels)
        return [];
    return [...labels].map((label) => ({
        textContent: label.textContent ?? undefined,
    }));
};
const parseNodeMap = (nodeMap) => {
    return Array.from(nodeMap).reduce((acc, attr) => {
        acc[attr.name] = attr.value;
        return acc;
    }, {});
};
const parseElement = (el) => {
    const base = {
        // adding a bunch of fields that are not on _all_ elements, but are on enough that it's useful to have them here.
        attributes: parseNodeMap(el.attributes),
        classList: [...el.classList],
        id: el.id,
        labels: parseLabels(el.labels),
        name: el.name,
        nodeName: el.nodeName,
        nodeValue: el.nodeValue,
        tagName: el.tagName,
        title: el.title,
        type: el.type,
        value: el.value,
    };
    if (el instanceof HTMLSelectElement) {
        return {
            ...base,
            selectedOptions: [...el.selectedOptions].map((option) => ({
                value: option.value,
                text: option.text,
            })),
            selectedIndex: el.selectedIndex,
        };
    }
    else if (el instanceof HTMLInputElement) {
        return {
            ...base,
            checked: el.checked,
        };
    }
    else if (el instanceof HTMLMediaElement) {
        return {
            ...base,
            currentSrc: el.currentSrc,
            currentTime: el.currentTime,
            duration: el.duration,
            ended: el.ended,
            muted: el.muted,
            paused: el.paused,
            playbackRate: el.playbackRate,
            readyState: el.readyState,
            src: el.src,
            volume: el.volume,
        };
    }
    else if (el instanceof HTMLButtonElement) {
        return {
            ...base,
            innerText: el.innerText,
        };
    }
    return base;
};
export class ClickSignalsGenerator {
    constructor() {
        this.id = 'click';
    }
    register(emitter) {
        const handleClick = (ev) => {
            const target = ev.target ?? {};
            if (this.isClickableElement(target)) {
                emitter.emit(createInteractionSignal({
                    eventType: 'click',
                    target: parseElement(target),
                }));
            }
        };
        document.addEventListener('click', handleClick, true);
        return () => document.removeEventListener('click', handleClick);
    }
    isClickableElement(el) {
        return (el instanceof HTMLAnchorElement ||
            el instanceof HTMLButtonElement ||
            ['button', 'link'].includes(el.getAttribute('role') ?? ''));
    }
}
export class FormSubmitGenerator {
    constructor() {
        this.id = 'form-submit';
    }
    register(emitter) {
        const handleSubmit = (ev) => {
            const target = ev.submitter;
            emitter.emit(createInteractionSignal({
                eventType: 'submit',
                submitter: parseElement(target),
            }));
        };
        document.addEventListener('submit', handleSubmit, true);
        return () => document.removeEventListener('submit', handleSubmit);
    }
}
export class OnChangeGenerator {
    constructor() {
        this.id = 'change';
    }
    register(emitter) {
        const handleChange = (ev) => {
            const target = ev.target;
            if (target instanceof HTMLInputElement) {
                if (target.type === 'password') {
                    logger.debug('Ignoring change event for input', target);
                    return;
                }
            }
            emitter.emit(createInteractionSignal({
                eventType: 'change',
                target: parseElement(target),
            }));
        };
        document.addEventListener('change', handleChange, true);
        return () => document.removeEventListener('change', handleChange);
    }
}
export class OnNavigationEventGenerator {
    constructor() {
        this.id = 'navigation';
    }
    register(emitter) {
        // emit navigation signal on page load
        emitter.emit(createNavigationSignal({
            action: 'pageLoad',
            ...this.createCommonFields(),
        }));
        // emit a navigation signal whenever the URL has changed
        const urlChange = new URLChangeObservable();
        urlChange.subscribe((prevUrl) => emitter.emit(createNavigationSignal({
            action: 'urlChange',
            prevUrl,
            ...this.createCommonFields(),
        })));
        return () => {
            urlChange.unsubscribe();
        };
    }
    createCommonFields() {
        return {
            // these fields are named after those from the page call, rather than a DOM api.
            url: location.href,
            path: location.pathname,
            hash: location.hash,
            search: location.search,
            title: document.title,
        };
    }
}
export const domGenerators = [
    ClickSignalsGenerator,
    FormSubmitGenerator,
    OnChangeGenerator,
    OnNavigationEventGenerator,
];
//# sourceMappingURL=dom-gen.js.map